package dev.pfaff.altargui.shader;

import dev.pfaff.glass.opengl.enums.GLVertexAttribPointerType;
import dev.pfaff.glass.opengl.enums.IGLVertexAttribPointerType;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.MemoryLayout;
import java.lang.foreign.ValueLayout;
import java.util.Objects;

import static java.lang.foreign.MemoryLayout.sequenceLayout;
import static java.lang.foreign.ValueLayout.JAVA_BYTE;

public sealed interface VertexComponentSpec {
	String name();

	MemoryLayout layout();

	record Padding(int byteSize) implements VertexComponentSpec {
		public Padding {
			if (byteSize <= 0) {
				throw new IllegalArgumentException(STR."byteSize must be greater than zero: \{byteSize}");
			}
		}

		@Override
		public String name() {
			return "padding";
		}

		@Override
		public MemoryLayout layout() {
			return this.byteSize == 1 ? JAVA_BYTE : sequenceLayout(this.byteSize, JAVA_BYTE);
		}
	}

	record Attrib(String name, IGLVertexAttribPointerType type, boolean normalized, int width, @Nullable MemoryLayout customLayout) implements VertexComponentSpec {
		public Attrib {
			if (normalized && !(type instanceof GLVertexAttribPointerType)) {
				throw new UnsupportedOperationException("Only GLVertexAttribPointerType can be normalized");
			}
			if (width <= 0) {
				throw new IllegalArgumentException(STR."width must be greater than zero: \{width}");
			}
			if (customLayout != null) {
				checkLayoutCompatibility(type, type.layout(), width, customLayout);
			}
		}

		public Attrib(String name, IGLVertexAttribPointerType type, boolean normalized, int width) {
			this(name, type, normalized, width, null);
		}

		public Attrib(String name, IGLVertexAttribPointerType type, int width) {
			this(name, type, false, width);
		}

		@Override
		public MemoryLayout layout() {
			if (this.customLayout != null) return this.customLayout;
			ValueLayout typeLayout = this.type.layout();
			if (this.width == 1) return typeLayout;
			return sequenceLayout(this.width, typeLayout);
		}

		public Attrib withNormalized() {
			return new Attrib(name, type, true, width, customLayout);
		}

		public Attrib withCustomLayout(MemoryLayout layout) {
			Objects.requireNonNull(layout);
			if (customLayout != null) {
				checkLayoutCompatibility(type, customLayout, 1, layout);
			}
			return new Attrib(name, type, normalized, width, layout);
		}

		private static void checkLayoutCompatibility(IGLVertexAttribPointerType type,
													 MemoryLayout base,
													 int baseWidth,
													 MemoryLayout custom) {
			if (custom.byteSize() != base.byteSize() * baseWidth || custom.byteAlignment() < base.byteAlignment()) {
				throw new IllegalArgumentException(STR."""
				Incompatible custom layout for \{type} attribute:
					Base layout: \{base}
					Custom layout: \{custom}
				""");
			}
		}
	}
}
